# Copyright 2017-2018 Tom Eulenfeld, GPLv3
"""Commands used by the CLI interface"""
import functools
import glob
import logging
import multiprocessing
import os
import re
import shutil
import textwrap

import h5py
import numpy as np
import obspy
from obspy.core import UTCDateTime as UTC
import obspyh5
import tqdm

import yam
from yam.correlate import correlate
from yam.io import _get_existent, _iter_h5, _write_corr, read_dicts, write_dict
import yam.stack
import yam.stretch
from yam.util import _analyze_key, _filter, _get_fname, IterTime, ParseError


log = logging.getLogger('yam.commands')




def _todo_tasks(tasks, done_tasks):
    if len(tasks) == 0:
        log.warning('no tasks found -> nothing to do')
    new_tasks = sorted(set(tasks) - set(done_tasks))
    if len(new_tasks) < len(tasks):
        msg = '%d of %d tasks already processed -> skip these tasks'
        log.info(msg, len(tasks) - len(new_tasks), len(tasks))
    return sorted(new_tasks)


def start_correlate(io,
                    filter_inventory=None,
                    startdate='1990-01-01', enddate='2020-01-01',
                    njobs=None,
                    parallel_inner_loop=False,
                    dtype='float16',
                    keep_correlations=False,
                    stack='1d',
                    **kwargs):
    """
    Start correlation

    :param io: |io|
    :param filter_inventory: filter inventory with its select method,
        specified dict is passed to |Inventory.select|
    :param str startdate,enddate: start and end date as strings
    : param njobs: number of cores to use for computation, days are computed
        parallel, this might consume much memory, default: None -- use all
        available cores
    :param parallel_inner_loop: Run inner loops parallel instead of outer loop
        (preproccessing of different stations and correlation of different
        pairs versus processing of different days).
        Useful for a datset with many stations.
    :param dtype: data type for storing correlations
        (default: float16 - half precision)
    :param keep_correlations,stack,\*\*kwargs: all other kwargs are passed to
        `~yam.correlate.correlate()` function
    """
    if filter_inventory:
        log.debug('filter inventory')
        io['inventory'] = io['inventory'].select(**filter_inventory)
    log.info('start preprocessing and correlation')
    tasks = [str(t)[:10] for t in IterTime(UTC(startdate), UTC(enddate))]
    done_tasks = None
    if stack is not None:
        key2 = kwargs['outkey'] + '_s' + stack
        done_tasks = [t[-16:-6] for t in _get_existent(io['stack'], key2, 4)]
    if keep_correlations:
        key2 = kwargs['outkey']
        done_tasks2 = [t[-16:-6] for t in _get_existent(io['corr'], key2, 4)]
        if done_tasks is None:
            done_tasks = done_tasks2
        else:
            done_tasks = [t for t in done_tasks if t in done_tasks2]
    tasks = _todo_tasks(tasks, done_tasks)
    tasks = [UTC(t) for t in tasks]
    kwargs.update({'keep_correlations': keep_correlations, 'stack': stack})
    if parallel_inner_loop:
        kwargs['njobs'] = njobs
        njobs = 1
    do_work = functools.partial(correlate, io, **kwargs)
    if njobs == 1:
        log.info('do work sequentially')
        for task in tqdm.tqdm(tasks, total=len(tasks)):
            result = do_work(task)
            _write_corr(result, io, dtype=dtype)
    else:
        pool = multiprocessing.Pool(njobs)
        log.info('do work parallel (%d cores)', pool._processes)
        for result in tqdm.tqdm(pool.imap_unordered(do_work, tasks),
                                total=len(tasks)):
            _write_corr(result, io, dtype=dtype)
        pool.close()
        pool.join()

    log.info('finished preprocessing and correlation')


def _stack_wrapper(groupnames, fname, outkey, **kwargs):
    """
    Wrapper around `~yam.stack.stack()`

    :param groupnames: groups to load the correlations from
    :param fname: file to load correlations from
    :param outkey: key to write stacked correlations to
    :param \*\*kwargs: all other kwargs are passed to
        `~yam.stack.stack()` function
    """
    with h5py.File(fname, 'r') as f:
        traces = [obspyh5.dataset2trace(f[g]) for g in groupnames]
    stream = obspy.Stream(traces)
    stack_stream = yam.stack.stack(stream, **kwargs)
    for tr in stack_stream:
        tr.stats.key = outkey
    return stack_stream


def start_stack(io, key, outkey, subkey='', njobs=None,
                starttime=None, endtime=None, **kwargs):
    """
    Start stacking

    :param io: |io|
    :param key:  key to load correlations from
    :param outkey: key to write stacked correlations to
    :param subkey: only use a part of the correlations
    :param njobs: number of cores to use for computation,
        default: None -- use all available cores
    :param starttime,endtime: constrain start and end dates
    :param \*\*kwargs: all other kwargs are passed to
        `yam.stack.stack()` function
    """
    fname = io['stack'] if 's' in _analyze_key(key) else io['corr']
    tasks = _get_existent(fname, key + subkey, 3)
    done_tasks = [t.replace(outkey, key) for t in
                  _get_existent(io['stack'], outkey + subkey, 3)]
    tasks = _todo_tasks(tasks, done_tasks)
    length = kwargs.get('length')
    for task in tqdm.tqdm(tasks, total=len(tasks)):
        subtasks = [t for t in _get_existent(fname, task, 4) if
                    (starttime is None or t[-16:] >= starttime) and
                    (endtime is None or t[-16:] <= endtime)]
        if length is None and njobs != 1:
            step = 1000
            subtask_chunks = [tuple(subtasks[i:i + step]) for i in
                              range(0, len(subtasks), step)]
        else:
            subtask_chunks = [subtasks]
            # TODO: parallel stacking for arbitrary stack id
#            lensec = _time2sec(length)
#            if lensec >= 30 * 24 * 3600:
#                subtask_chunks = [subtasks]
#            else:
#                subtask_chunks = []
#                for i in range(0, len(subtasks), step):
#                    chunk = subtasks[i:i + step]
#                    t1 = UTC(subtasks[i + step - 1][-16:])
#                    j = 0
#                    while i + step + j < len(subtasks):
#                        t2 = UTC(subtasks[i + step + j][-16:])
#                        # assume lensec is always larger than movesec
#                        # not ideal, may load to much data
#                        # eg for stack over 1 year
#                        if t2 - t1 <= lensec:
#                            chunk.append(subtasks[i + step + j])
#                        else:
#                            break
#                        j += 1
#                    subtask_chunks.append(chunk)
        do_work = functools.partial(_stack_wrapper, fname=fname, outkey=outkey,
                                    **kwargs)
        results = []
        if njobs == 1 or len(subtask_chunks) == 1:
            log.debug('do work sequentially')
            for stask in tqdm.tqdm(subtask_chunks, total=len(subtask_chunks)):
                result = do_work(stask)
                results.append(result)
        else:
            pool = multiprocessing.Pool(njobs)
            log.debug('do work parallel (%d cores)', pool._processes)
            for result in tqdm.tqdm(
                    pool.imap(do_work, subtask_chunks),
                    total=len(subtask_chunks)):
                results.append(result)
            pool.close()
            pool.join()
        if length is None:
            for stream in results:
                assert len(stream) <= 1
            traces = [tr for stream in results for tr in stream]
            num = sum(tr.stats.num for tr in traces)
            data = np.sum([tr.data * (tr.stats.num / num) for tr in traces],
                          axis=0)
            tr_stack = obspy.Trace(data, header=traces[0].stats)
            tr_stack.stats.num = num
            tr_stack.write(io['stack'], 'H5', mode='a')
        else:
            for stack_stream in results:
                stack_stream.write(io['stack'], 'H5', mode='a')


def _stretch_wrapper(groupnames, fname, outkey, filter=None,
                     **kwargs):
    """
    Wrapper around `~yam.stretch.stretch()`

    :param groupname: group to load the correlations from
    :param fname: file to load correlations from
    :param fname_stretch: file for writing results
    :param outkey: key to write stretch results to
    :param filter: filter correlations before stretching
        (bandpass, tuple with min and max frequency)
    :param \*\*kwargs: all other kwargs are passed to
        `~yam.stretch.stretch()` function
    """
    with h5py.File(fname, 'r') as f:
        traces = [obspyh5.dataset2trace(f[g]) for g in groupnames]
    stream = obspy.Stream(traces)
    for tr in stream:
        tr.data = np.require(tr.data, float)
    if filter:
        _filter(stream, filter)
    stretchres = yam.stretch.stretch(stream, **kwargs)
    if stretchres is not None:
        stretchres['attrs']['key'] = outkey
        return stretchres


def start_stretch(io, key, subkey='', njobs=None, reftrid=None,
                  starttime=None, endtime=None, dtype='float16',
                  **kwargs):
    """
    Start stretching

    :param io: |io|
    :param key:  key to load correlations from
    :param subkey: only use a part of the correlations
    :param njobs: number of cores to use for computation,
        default: None -- use all available cores
    :param reftrid: Parallel processing is only possible when this parameter
        is specified. Key to load the reference trace from, e.g. `'c1_s'`,
        it can be created by a command similar to `yam stack c1 ''`.
    :param starttime,endtime: constrain start and end dates
    :param dtype: data type for storing similarity matrices
        (default: float16 - half precision)
    :param \*\*kwargs: all other kwargs are passed to
        `stretch_wrapper()` function
    """
    fname = _get_fname(io, key)
    outkey = kwargs['outkey']
    tasks = _get_existent(fname, key + subkey, 3)
    done_tasks = [t.replace(outkey, key) for t in
                  _get_existent(io['stretch'], outkey + subkey, 3)]
    tasks = _todo_tasks(tasks, done_tasks)
    for task in tqdm.tqdm(tasks, total=len(tasks)):
        if reftrid is None:
            reftr = None
        else:
            fname_reftr = _get_fname(io, reftrid)
            group_reftr = task.replace(key, reftrid)
            reftr = obspy.read(fname_reftr, 'H5', group=group_reftr,
                               dtype=float)
            if len(reftr) != 1:
                raise NotImplementedError('Reference must be single trace')
            reftr = reftr[0]
        subtasks = [t for t in _get_existent(fname, task, 4) if
                    (starttime is None or t[-16:] >= starttime) and
                    (endtime is None or t[-16:] <= endtime)]
        if reftr is None:
            subtask_chunks = [tuple(subtasks)]
        else:
            step = 1000
            subtask_chunks = [tuple(subtasks[i:i + step]) for i in
                              range(0, len(subtasks), step)]
        do_work = functools.partial(_stretch_wrapper, fname=fname,
                                    reftr=reftr, **kwargs)
        results = []
        if njobs == 1 or len(subtask_chunks) == 1:
            log.debug('do work sequentially')
            for stask in tqdm.tqdm(subtask_chunks, total=len(subtask_chunks)):
                result = do_work(stask)
                results.append(result)
        else:
            pool = multiprocessing.Pool(njobs)
            log.debug('do work parallel (%d cores)', pool._processes)
            for result in tqdm.tqdm(
                    pool.imap(do_work, subtask_chunks),
                    total=len(subtask_chunks)):
                results.append(result)
            pool.close()
            pool.join()
        result = yam.stretch.join_dicts(results)
        if result is not None:
            write_dict(result, io['stretch'], dtype=dtype)


def _start_ipy(obj):
    from IPython import start_ipython
    print('Contents loaded into obj variable.')
    start_ipython(argv=[], user_ns={'obj': obj}, display_banner=False)
    print('Good Bye')


def _get_print2():
    num, _ = shutil.get_terminal_size()
    if num == 0:
        num = 80
    wrap = textwrap.TextWrapper(width=num - 1, initial_indent='    ',
                                subsequent_indent='    ')

    def print2(text):
        print(wrap.fill(text))
    return print2


def _get_data_glob(data):
    """
    Construct a glob expression from the data expression
    """
    return re.sub(r'{[^{}]*}', '*', data)


def _get_data_files(data):
    """
    Construct a glob expression from the data expression and return file names
    """
    return glob.glob(_get_data_glob(data))


def _print_info_helper(key, io):
    print2 = _get_print2()
    is_stretch = key == 'tstretch'
    fname = _get_fname(io, key)
    keys = _get_existent(fname, '/', 1)  # 1, 3, 4
    if len(keys) == 0:
        print2('None')
    for key in sorted(keys):
        keys2 = _get_existent(fname, key, 3)
        subkey = key.split('/')[-1]
        if is_stretch:
            o = '%s: %d combs' % (subkey, len(keys2))
        else:
            keys3 = _get_existent(fname, key, 4)
            o = ('%s: %d combs, %d corrs' %
                 (subkey, len(keys2), len(keys3)))
        print2(o)


def info(io, key=None, subkey='', config=None, **unused_kwargs):
    """
    Print information about yam project

    :param io: |io|
    :param key: key to print infos about
        (key inside HDF5 file, or one of data, stations,
        default: None -- print overview)
    :param subkey: only print part of the HDF5 file
    :param config: list of configuration dictionaries
    """
    print2 = _get_print2()
    data_plugin = io.get('data_plugin')
    if key is None:
        print('Stations:')
        inventory = io['inventory']
        if inventory is None:
            print2('Not found')
        else:
            stations = inventory.get_contents()['stations']
            channels = inventory.get_contents()['channels']
            print2(' '.join(st.strip().split()[0] for st in stations))
            print2('%d stations, %d channels' % (len(stations), len(channels)))
        if data_plugin:
            print('Data plugin:')
            print2('%s' % data_plugin)
        else:
            print('Raw data (expression for day files):')
            print2(io['data'])
            print2('%d files found' % len(_get_data_files(io['data'])))
        print('Config ids:')

        def get_keys(d):
            if d is None or len(d) == 0:
                return 'None'
            else:
                return ', '.join(sorted(d.keys()))
        print2('c Corr: ' + get_keys(config[0]))
        print2('s Stack: ' + get_keys(config[1]))
        print2('t Stretch: ' + get_keys(config[2]))
        print('Correlations (channel combinations, correlations calculated):')
        _print_info_helper('corr', io)
        print('Stacks:')
        _print_info_helper('stack', io)
        print('Stretching matrices:')
        _print_info_helper('tstretch', io)
    elif key == 'stations':
        print(io['inventory'])
    elif key == 'data':
        if data_plugin:
            print('Data plugin:')
            print2('%s' % data_plugin)
        else:
            print('Raw data (expression for day files):')
            print2(io['data'])
            fnames = _get_data_files(io['data'])
            print2('%d files found' % len(fnames))
            for fname in sorted(fnames):
                print2(fname)
    else:
        is_stretch = 't' in _analyze_key(key)
        fname = _get_fname(io, key)
        level = 3 if is_stretch else 4
        for line in _get_existent(fname, key + subkey, level):
            print2(line)


def _load_data(seedid, day, data, data_format, key='data',
               **prep_kw):
    """Load preprocessed or raw data"""

    from obspy import UTCDateTime as UTC
    from yam.util import _seedid2meta
    from yam.correlate import get_data, preprocess
    smeta = _seedid2meta(seedid)
    day = UTC(day)
    if key == 'data':
        obj = get_data(smeta, data, data_format, day,
                       overlap=0, edge=0, trim_and_merge=True)
        return obj
    stream = get_data(smeta, data, data_format, day,
                      overlap=0, edge=60, trim_and_merge=False)
    preprocess(stream, day, **prep_kw)
    return stream


def load(io, key, seedid=None, day=None, do='return', prep_kw={},
         fname=None, format=None):
    """
    Load object and do something with it

    :param io: io
    :param key: key of object to load
        (key inside HDF5 file, or one of data, prepdata, stations)
    :param seedid: seed id of a  channel (for data or prepdata)
    :param day: |UTC| object with day (for data or prepdata)
    :param do: specifies what to do with the object, default is ``'return'``
        which simply returns the object, other possible values are
        ``'print'`` -- print object (used by print command),
        ``'load'`` -- load object in IPython session (used by load command),
        ``'export'`` -- export correlations to different file format
        (used by export command)
    :param dict prep_kw: options passed to preprocess (for prepdata only)
    :param fname: file name (for export command)
    :param format: target format (for export command)
    """
    if key == 'stations':
        obj = io['inventory']
    elif key in ('data', 'prepdata'):
        if seedid is None or day is None:
            msg = 'seedid and day must be given for data or prepdata'
            raise ParseError(msg)
        if key == 'prepdata':
            prep_keys = ('remove_response', 'remove_response_options',
                         'demean', 'filter', 'normalization',
                         'time_norm_options', 'spectral_whitening_options',
                         'tolerance_shift',
                         'downsample')
            prep_kw = {k: prep_kw.get(k) for k in prep_keys}
        obj = _load_data(seedid, day, io['data'], io.get('data_format'),
                         key, inventory=io['inventory'], **prep_kw)
    else:
        is_stretch = 't' in _analyze_key(key)
        fname_in = _get_fname(io, key)
        if is_stretch:
            obj = read_dicts(fname_in, key)
            if do == 'print':
                obj = '\n\n'.join(str(o) for o in obj)
        else:
            obj = obspy.read(fname_in, 'H5', group=key, headonly=do == 'print')
            if do == 'print':
                obj = obj.__str__(extended=True)
    if do == 'print':
        print(obj)
    elif do == 'load':
        _start_ipy(obj)
    elif do == 'return':
        return obj
    elif do == 'export':
        print('export', obj)
        if format == 'H5':
            obspyh5.set_index()
        obj.write(fname, format)
        if format == 'H5':
            from yam.io import INDEX
            obspyh5.set_index(INDEX)
    else:
        raise


def plot(io, key, plottype=None, seedid=None, day=None, prep_kw={},
         corrid=None, show=False,
         **kwargs):
    """
    Plot everything

    :param io: |io|
    :param key: key of objects to plot, or one of stations, data, prepdata
    :param plottype: plot type to use
        (non default values are ``'vs_dist'`` and ``'wiggle'`` for
        correlation plots)
    :param seedid: seed id of a  channel (for data or prepdata)
    :param day: |UTC| object with day (for data or prepdata)
    :param dict prep_kw: options passed to preprocess (for prepdata only)
    :param corrid: correlation configuration (for prepdata only)
    :param show: show interactive plot
    :param \*\*kwargs: all other kwargs are passed to
        the corresponding plot function in `~yam.imaging` module
    """
    import yam.imaging
    path = io['plot']
    if not os.path.exists(path):
        os.mkdir(path)
    if key in ('stations', 'data', 'prepdata'):
        pt = key
    else:
        is_corr = 't' not in _analyze_key(key)
        if is_corr and plottype == 'vs_dist':
            pt = 'corr_vs_dist'
        elif is_corr and plottype == 'wiggle':
            pt = 'corr_vs_time_wiggle'
        elif is_corr and plottype is None:
            pt = 'corr_vs_time'
        elif not is_corr and plottype is None:
            pt = 'sim_mat'
        else:
            raise ParseError('Combination of key and plottype not supported')

    kw = kwargs.get('plot_%s_options' % pt, {})
    kw.update(kwargs.get('plot_options', {}))
    bname = os.path.join(path, pt)
    if key == 'stations':
        yam.imaging.plot_stations(io['inventory'], bname, **kw)
    elif key in ('data', 'prepdata'):
        data = load(io, key, do='return', seedid=seedid, day=day,
                    prep_kw=prep_kw)
        fname = bname + '_%s_%s' % (seedid, day)
        if key == 'prepdata':
            fname = fname + '_c' + corrid
        yam.imaging.plot_data(data, fname, show=show, **kw)
    else:
        plot_ = getattr(yam.imaging, 'plot_' + pt)
        if pt == 'corr_vs_dist':
            fname2 = _get_fname(io, key)
            stream = obspy.read(fname2, 'H5', group=key)
            fname = bname + '_' + key.replace('/', '_')
            plot_(stream, fname, **kw)
        else:
            for task, res in _iter_h5(io, key):
                fname = bname + task.replace('/', '_')
                plot_(res, fname, **kw)
    if show:
        from matplotlib.pyplot import show
        show()


def remove(io, keys):
    """
    Remove one or several keys from HDF5 file

    :param io: |io|
    :param keys: list of keys to remove
    """
    for key in keys:
        if '/' in key and key.split('/', 1) != '':
            from warnings import warn
            warn('It is highly encouraged to delete only top level keys')
        fname = _get_fname(io, key)
        with h5py.File(fname, 'a') as f:
            del f[key]



# Copyright 2017-2019 Tom Eulenfeld, MIT license
"""Preprocessing and correlation"""
from functools import partial
import itertools
import logging
import multiprocessing

import numpy as np
from numpy.fft import rfft, irfft, rfftfreq
import obspy
from obspy.core import Stream
from obspy.geodetics import gps2dist_azimuth
from obspy.signal.cross_correlation import correlate as obscorr
from scipy.fftpack import fft, ifft, fftshift, ifftshift, next_fast_len
from scipy.signal import freqz, iirfilter, hilbert

from yam.util import _filter, IterTime, smooth as smooth_func, _time2sec
import yam.stack


log = logging.getLogger('yam.correlate')


def start_parallel_jobs_inner_loop(tasks, do_work, njobs=1):
    if njobs == 1:
        results = [do_work(task) for task in tasks]
    else:
        pool = multiprocessing.Pool(njobs)
        results = pool.map(do_work, tasks)
        pool.close()
        pool.join()
    return results


def _fill_array(data, mask=None, fill_value=None):
    """
    Mask numpy array and/or fill array value without demasking.

    Additionally set fill_value to value.
    If data is not a MaskedArray and mask is None returns silently data.

    :param mask: apply mask to array
    :param fill_value: fill value
    """
    if mask is not None and mask is not False:
        data = np.ma.MaskedArray(data, mask=mask, copy=False)
    if np.ma.is_masked(data) and fill_value is not None:
        data._data[data.mask] = fill_value
        np.ma.set_fill_value(data, fill_value)
#    elif not np.ma.is_masked(data):
#        data = np.ma.filled(data)
    return data


def time_norm(tr, method,
              clip_factor=None, clip_set_zero=None,
              clip_value=2, clip_std=True, clip_mode='clip',
              mute_parts=48, mute_factor=2, plugin=None,
              plugin_options={}):
    """
    Calculate normalized data, see e.g. Bensen et al. (2007)

    :param tr: Trace to manipulate
    :param str method:
        1bit: reduce data to +1 if >0 and -1 if <0\n
        clip: clip data to value or multiple of root mean square (rms)\n
        mute_envelope: calculate envelope and set data to zero where envelope
        is larger than specified
        plugin: use own function
    :param mask_zeros: mask values that are set to zero, they will stay zero
        in the further processing
    :param float clip_value: value for clipping or list of lower and upper
        value
    :param bool clip_std: Multiply clip_value with rms of data
    :param bool clip_mode:
        'clip': clip data
        'zero': set clipped data to zero
        'mask': set clipped data to zero and mask it
    :param int mute_parts: mean of the envelope is calculated by dividing the
        envelope into several parts, the mean calculated in each part and
        the median of this averages defines the mean envelope
    :param float mute_factor: mean of envelope multiplied by this
        factor defines the level for muting
    :param str plugin: function in the form module:func
    :param dict plugin_options: kwargs passed to plugin

    :return: normalized data
    """
    data = tr.data
    data = _fill_array(data, fill_value=0)
    mask = np.ma.getmask(data)
    if method == '1bit':
        np.sign(data, out=data)
    elif method == 'clip':
        if clip_factor is not None:
            from warnings import warn
            msg = 'clip_factor is deprecated, use clip_value instead'
            warn(msg, DeprecationWarning)
            clip_value = clip_factor
        if clip_set_zero is not None:
            from warnings import warn
            msg = 'clip_set_zero is deprecated, use clip_mode instead'
            warn(msg, DeprecationWarning)
            clip_mode = 'zero' if clip_set_zero else 'clip'
        from collections import Iterable
        if not isinstance(clip_value, Iterable):
            clip_value = [-clip_value, clip_value]
        if clip_std:
            std = np.std(data)
            clip_value = [clip_value[0] * std, clip_value[1] * std]
        if clip_mode == 'clip':
            np.clip(data, *clip_value, out=data)
        else:
            cmask = np.logical_or(data < clip_value[0], data > clip_value[1])
            if clip_mode == 'mask':
                mask = np.logical_or(np.ma.getmaskarray(data), cmask)
            elif clip_mode == 'zero':
                data[cmask] = 0
            else:
                raise ValueError('clip_mode must be one of clip, zeros, mask')
    elif method == 'mute_envelope':
        N = next_fast_len(len(data))
        envelope = np.abs(hilbert(data, N))[:len(data)]
        levels = [np.mean(d) for d in np.array_split(envelope, mute_parts)]
        level = mute_factor * np.median(levels)
        data[envelope > level] = 0
    elif method == 'plugin':
        from yam.util import _load_func
        modulename, funcname = plugin.split(':')
        func = _load_func(modulename.strip(), funcname.strip())
        func(tr, **plugin_options)
        data = tr.data
    else:
        msg = 'The method passed to time_norm is not known: %s.' % method
        raise ValueError(msg)
    tr.data = _fill_array(data, mask=mask, fill_value=0)
    return tr


# http://azitech.wordpress.com/
# 2011/03/15/designing-a-butterworth-low-pass-filter-with-scipy/
def _filter_resp(freqmin, freqmax, corners=2, zerophase=False, sr=None,
                 N=None, whole=False):
    """
    Complex frequency response of Butterworth-Bandpass Filter.

    :param freqmin: Pass band low corner frequency.
    :param freqmax: Pass band high corner frequency.
    :param corners: Filter corners
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :param sr: Sampling rate in Hz.
    :param N,whole: passed to scipy.signal.freqz

    :return: frequencies and complex response
    """
    df = sr
    fe = 0.5 * df
    low = freqmin / fe
    high = freqmax / fe
    # raise for some bad scenarios
    if high > 1:
        high = 1.0
        msg = "Selected high corner frequency is above Nyquist. " + \
              "Setting Nyquist as high corner."
        log.warning(msg)
    if low > 1:
        msg = "Selected low corner frequency is above Nyquist."
        raise ValueError(msg)
    [b, a] = iirfilter(corners, [low, high], btype='band',
                       ftype='butter', output='ba')
    freqs, values = freqz(b, a, N, whole=whole)
    if zerophase:
        values *= np.conjugate(values)
    return freqs, values


def spectral_whitening(tr, smooth=None, filter=None,
                       waterlevel=1e-8, mask_again=True):
    """
    Apply spectral whitening to data

    Data is divided by its smoothed (Default: None) amplitude spectrum.

    :param tr: trace to manipulate
    :param smooth: length of smoothing window in Hz
        (default None -> no smoothing)
    :param filter: filter spectrum with bandpass after whitening
        (tuple with min and max frequency)
    :param waterlevel: waterlevel relative to mean of spectrum
    :param mask_again: weather to mask array after this operation again and
        set the corresponding data to 0

    :return: whitened data
    """
    sr = tr.stats.sampling_rate
    data = tr.data
    data = _fill_array(data, fill_value=0)
    mask = np.ma.getmask(data)
    nfft = next_fast_len(len(data))
    spec = fft(data, nfft)
    spec_ampl = np.abs(spec)
    spec_ampl /= np.max(spec_ampl)
    if smooth:
        smooth = int(smooth * nfft / sr)
        spec_ampl = ifftshift(smooth_func(fftshift(spec_ampl), smooth))
    # save guard against division by 0
    spec_ampl[spec_ampl < waterlevel] = waterlevel
    spec /= spec_ampl
    if filter is not None:
        spec *= _filter_resp(*filter, sr=sr, N=len(spec), whole=True)[1]
    ret = np.real(ifft(spec, nfft)[:len(data)])
    if mask_again:
        ret = _fill_array(ret, mask=mask, fill_value=0)
    tr.data = ret
    return tr


def __get_stations(inventory):
    channels = inventory.get_contents()['channels']
    stations = sorted({ch[:-1] + '?': ch[-1] for ch in channels})
    return stations


def _iter_station_meta(inventory, components):
    """
    Return iterator yielding metadata per station and day.

    :param inventory: |Inventory| object with station and channel information
    :param components: components to yield
    """
    stations = __get_stations(inventory)
    for seedid in stations:
        for comp in components:
            net, sta, loc, cha = seedid.split('.')
            cha = cha[:2] + comp
            meta = {'network': net, 'station': sta, 'location': loc,
                    'channel': cha}
            yield meta


def get_data(smeta, data, data_format, day, overlap=0, edge=0,
             trim_and_merge=False):
    """Return data of one day

    :param smeta: dictionary with station metadata
    :param data: string with expression of data day files or
        function that returns the data (aka get_waveforms)
    :param data_format: format of data
    :param day: day as |UTC| object
    :param overlap: overlap to next day in seconds
    :param edge: additional time span requested from day before and after
        in seconds
    :param trim_and_merge: weather data is trimmed to day boundaries and merged
    """
    next_day = day + 24 * 3600
    if not isinstance(data, str):
        try:
            stream = data(starttime=day - edge,
                          endtime=next_day + overlap + edge, **smeta)
        except Exception as ex:
            log.debug('no data for %s %s: %s', day, smeta, str(ex))
            return
    else:
        fname = data.format(t=day, **smeta)
        try:
            stream = obspy.read(fname, data_format)
        except:
            return
        t1 = stream[0].stats.starttime
        t2 = stream[-1].stats.endtime
        if t1 - day < 60:
            fname = data.format(t=day - 1, **smeta)
            try:
                stream += obspy.read(fname, data_format, starttime=day - edge)
            except:
                pass
        if next_day - t2 < 60:
            endtime = next_day + overlap + edge
            fname = data.format(t=next_day, **smeta)
            try:
                stream += obspy.read(fname, data_format, endtime=endtime)
            except:
                pass
    if trim_and_merge:
        stream.merge(method=1, interpolation_samples=10)
        stream.trim(day, next_day + overlap)
    return stream


def _shift(trace, shift):
    """Shift trace by given time and correct starttime => interpolation"""
    msg = ('interpolate trace %s with starttime %s to shift by %.6fs '
           '(Fourier method)')
    log.debug(msg, trace.id, trace.stats.starttime, shift)
    nfft = next_fast_len(len(trace))
    spec = rfft(trace.data, nfft)
    freq = rfftfreq(nfft, trace.stats.delta)
    spec *= np.exp(-2j * np.pi * freq * shift)
    trace.data = irfft(spec, nfft)[:len(trace)]
    trace.stats.starttime -= shift
    return trace


def _downsample_and_shift(trace, target_sr=None, tolerance_shift=None,
                          **interpolate_options):
    """Downsample and align samples at "good" times by shifting"""
    sr = trace.stats.sampling_rate
    if target_sr is None:
        target_sr = sr
    dt = 1 / target_sr
    shift = (1e-6 * trace.stats.starttime.microsecond) % dt
    if shift > 0.5 * dt:
        shift = shift - dt
    if tolerance_shift is None:
        tolerance_shift = np.finfo(float).eps
    must_shift = abs(shift) > tolerance_shift
    if not must_shift:  # anyway correct starttime
        trace.stats.starttime -= shift
    if sr % target_sr == 0:
        if sr != target_sr:
            trace.decimate(int(sr // target_sr))
        if must_shift:
            _shift(trace, shift)
    else:
        # anti-aliasing filter
        if sr / target_sr > 16:
            msg = ('Automatic filter design is unstable for decimation'
                   ' factors above 16. '
                   'Manual decimation is necessary.')
            raise ArithmeticError(msg)
        trace.filter('lowpass_cheby_2', freq=0.5 * target_sr,
                     maxorder=12)
        if must_shift:
            starttime = trace.stats.starttime - shift
            if starttime < trace.stats.starttime:
                starttime += dt
            msg = ('interpolate trace %s with starttime %s to downsample and '
                   'shift by %.6fs (Stream.interpolate() method)')
            log.debug(msg, trace.id, trace.stats.starttime, shift)
        else:
            starttime = None
        trace.interpolate(target_sr, starttime=starttime,
                          **interpolate_options)
    return trace


def _prep1(target_sr, tolerance_shift, interpolate_options,
           remove_response, inventory, remove_response_options, demean, filter,
           tr):
    """Helper function for parallel preprocessing"""
    tr.data = tr.data.astype('float64')
    _downsample_and_shift(tr, target_sr=target_sr,
                          tolerance_shift=tolerance_shift,
                          interpolate_options=interpolate_options)
    if remove_response:
        tr.remove_response(inventory, **remove_response_options)
    if demean:
         tr.detrend('demean')
    if filter is not None:
        _filter(tr, filter)
    return tr


def _prep2(normalization, time_norm_options, spectral_whitening_options,
           decimate,
           tr):
    """Helper function for parallel preprocessing"""
    tr.data = _fill_array(tr.data, fill_value=0)
    for norm in normalization:
        if norm == 'spectral_whitening':
            spectral_whitening(tr, **spectral_whitening_options)
        else:
            time_norm(tr, norm, **time_norm_options)
    if decimate:
        mask = np.ma.getmask(tr.data)
        tr.decimate(decimate, no_filter=True)
        if mask is not np.ma.nomask:
            tr.data = np.ma.MaskedArray(tr.data, mask[::decimate],
                                        fill_value=0)
    return tr


def preprocess(stream, day=None, inventory=None,
               overlap=0,
               remove_response=False,
               remove_response_options=None,
               demean=True,
               filter=None,
               normalization=(),
               time_norm_options=None,
               spectral_whitening_options=None,
               downsample=None, tolerance_shift=None, interpolate_options=None,
               decimate=None,
               njobs=1):
    """
    Preprocess stream of 1 day

    :param stream: |Stream| object
    :param day: |UTC| object of day (for trimming)
    :param inventory: |Inventory| object (for response removal)
    :param bool remove_response: remove response
    :param filter: min and max frequency of bandpass filter
    :param normalizaton: ordered list of normalizations to apply,
        ``'sprectal_whitening'`` for `spectral_whitening` and/or
        one or several of the time normalizations listed in `time_norm`
    :param downsample: downsample before preprocessing,
        target sampling rate
    :param tolerance_shift: Samples are aligned at "good" times for the target
        sampling rate. Specify tolerance in seconds. (default: no tolerance)
    :param decimate: decimate further by given factor after preprocessing
        (see Trace.decimate)
    :param njobs: number of parallel workers
    :param \*_options: dictionary of options passed to the corresponding
        functions
    """
    if time_norm_options is None:
        time_norm_options = {}
    if spectral_whitening_options is None:
        spectral_whitening_options = {}
    spectral_whitening_options.setdefault('filter', filter)
    if remove_response_options is None:
        remove_response_options = {}
    if interpolate_options is None:
        interpolate_options = {}
    if isinstance(normalization, str):
        normalization = [normalization]
    stream.merge(1, interpolation_samples=10)
    stream.traces = stream.split().traces
    # discard traces with less than 10 samples
    stream.traces = [tr for tr in stream if len(tr) >= 10]
    if downsample is None:
        downsample = min(tr.stats.sampling_rate for tr in stream)
    # call _prep1 on all traces, merge stream and call _prep2 on all traces
    do_work = partial(_prep1, downsample, tolerance_shift, interpolate_options,
                      remove_response, inventory, remove_response_options,
                      demean, filter)
    stream.traces = start_parallel_jobs_inner_loop(stream, do_work, njobs)
    len1 = len(stream)
    stream.merge()
    if len(stream) < len1:
        log.debug('detected gaps in data')
    do_work = partial(_prep2, normalization, time_norm_options,
                      spectral_whitening_options, decimate)
    stream.traces = start_parallel_jobs_inner_loop(stream, do_work, njobs)
    if day is not None:
        next_day = day + 24 * 3600
        stream.trim(day, next_day + overlap)
    stream.sort()
    assert len({tr.stats.sampling_rate for tr in stream}) == 1
    return stream


def correlate_traces(tr1, tr2, maxshift=3600, demean=True):
    """
    Return trace of cross-correlation of two input traces

    :param tr1,tr2: two |Trace| objects
    :param maxsift: maximal shift in correlation in seconds
    """
    n1, s1, l1, c1 = tr1.id.split('.')
    n2, s2, l2, c2 = tr2.id.split('.')
    sr = tr1.stats.sampling_rate
    xdata = obscorr(tr1.data, tr2.data, int(round(maxshift * sr)),
                    demean=demean)
    header = {'network': s1, 'station': c1, 'location': s2, 'channel': c2,
              'network1': n1, 'station1': s1, 'location1': l1, 'channel1': c1,
              'network2': n2, 'station2': s2, 'location2': l2, 'channel2': c2,
              'starttime': tr1.stats.starttime,
              'sampling_rate': sr,
              }
    return obspy.Trace(data=xdata, header=header)


def _make_same_length(tr1, tr2):
    """Guarantee that tr1 and tr2 have the same length.

    Even if tr1 and tr2 have the same sampling rate and are trimmed with
    the same times, they could differ in length up to one sample.
    This is handled here.
    """
    dlen = len(tr2) - len(tr1)
    dt = tr1.stats.delta
    if dlen == -1:
        tr1, tr2 = tr2, tr1
    if abs(dlen) == 1:
        # tr2 is too long
        if tr1.stats.starttime - tr2.stats.starttime > dt / 2:
            tr2.data = tr2.data[1:]
            tr2.stats.starttime += dt
        else:
            tr2.data = tr2.data[:-1]
    elif abs(dlen) > 1:
        msg = 'This should not happen ;), traces have different length'
        raise ValueError(msg)


def _slide_and_correlate_traces(day, next_day, length, overlap, discard,
                                max_lag, outkey, demean_window,
                                task):
    """Helper function for parallel correlating"""
    tr1, tr2, dist, azi, baz = task
    sr = tr1.stats.sampling_rate
    sr2 = tr2.stats.sampling_rate
    if sr != sr2:
        msg = 'Traces have different sampling rate (%s != %s)' % (sr, sr2)
        raise ValueError(msg)
    xstream = obspy.Stream()
    for t1 in IterTime(day, next_day - length + overlap, dt=length - overlap):
        sub = obspy.Stream([tr1, tr2]).slice(t1, t1 + length)
        if len(sub) < 2:
            continue
        st = [tr.stats.starttime for tr in sub]
        et = [tr.stats.endtime for tr in sub]
        if max(st) > min(et):  # this should not happen
            continue
        sub.trim(max(st), min(et))
        _make_same_length(sub[0], sub[1])
        avail = min((tr.data.count() if hasattr(tr.data, 'count')
                     else len(tr)) / sr / length for tr in sub)
        if discard is not None and avail < discard:
            msg = ('discard trace combination %s-%s for time %s '
                   '(availability %.1f%% < %.1f%% desired)')
            log.debug(msg, sub[0].id, sub[1].id, str(max(st))[:19],
                      100 * avail, 100 * discard)
            continue
        for tr in sub:
            _fill_array(tr.data, fill_value=0)
            tr.data = np.ma.getdata(tr.data)
        xtr = correlate_traces(sub[0], sub[1], max_lag, demean=demean_window)
        xtr.stats.starttime = t1
        xtr.stats.key = outkey
        xtr.stats.dist = dist
        xtr.stats.azi = azi
        xtr.stats.baz = baz
        xtr.stats.avail = avail
        xstream += xtr
    return xstream


def _midtime(stats):
    return stats.starttime + 0.5 * (stats.endtime - stats.starttime)


def correlate(io, day, outkey,
              edge=60,
              length=3600, overlap=1800,
              demean_window=True,
              discard=None,
              only_auto_correlation=False,
              station_combinations=None,
              component_combinations=('ZZ',),
              max_lag=100,
              keep_correlations=False,
              stack='1d',
              njobs=1,
              **preprocessing_kwargs):
    """
    Correlate data of one day

    :param io: io config dictionary
    :param day: |UTC| object with day
    :param outkey: the output key for the HDF5 index
    :param edge: additional time span requested from day before and after
        in seconds
    :param length: length of correlation in seconds (string possible)
    :param overlap: length of overlap in seconds (string possible)
    :param demean_window: demean each window individually before correlating
    :param discard: discard correlations with less data coverage
        (float from interval [0, 1])
    :param only_auto_correlations: Only correlate stations with itself
        (different components possible)
    :param station_combinations: specify station combinations
        (e.g. ``'CX.PATCX-CX.PB01``, network code can be
        omitted, e.g. ``'PATCX-PB01'``, default: all)
    :param component_combinations: component combinations to calculate,
        tuple of strings with length two, e.g. ``('ZZ', 'ZN', 'RR')``,
        if ``'R'`` or ``'T'`` is specified, components will be rotated after
        preprocessing, default: all component combinations (not rotated)
    :param max_lag: max time lag in correlations in seconds
    :param keep_correlatons: write correlations into HDF5 file (dafault: False)
    :param stack: stack correlations and write stacks into HDF5 file
        (default: ``'1d'``, must be smaller than one day or one day)

        .. note::

            If you want to stack larger time spans
            use the separate stack command on correlations or stacked
            correlations.

    :param njobs: number of jobs used. Some tasks will run parallel
        (preprocessing and correlation).
    :param \*\*preprocessing_kwargs: all other kwargs are passed to
        `preprocess`

    """
    inventory = io['inventory']
    length = _time2sec(length)
    overlap = _time2sec(overlap)
    if not keep_correlations and stack is None:
        msg = ('keep_correlation is False and stack is None -> correlations '
               ' would not be saved')
        raise ValueError(msg)
    components = set(''.join(component_combinations))
    if 'R' in components or 'T' in components:
        load_components = components - {'R', 'T'} | {'N', 'E'}
    else:
        load_components = components
    if station_combinations is not None:
        load_stations = set(sta for comb in station_combinations
                            for sta in comb.split('-'))
    else:
        load_stations = None
    # load data
    stream = obspy.Stream()
    for smeta in _iter_station_meta(inventory, load_components):
        if (load_stations is not None and smeta['station'] not in load_stations
                and '.'.join((smeta['network'], smeta['station']))
                not in load_stations):
            continue
        stream2 = get_data(smeta, io['data'], io['data_format'], day,
                           overlap=overlap, edge=edge)
        if stream2:
            stream += stream2
    if len(stream) == 0:
        log.warning('empty stream for day %s', str(day)[:10])
        return
    preprocess(stream, day, inventory, overlap=overlap, njobs=njobs,
               **preprocessing_kwargs)
    # collect trace pairs for correlation
    next_day = day + 24 * 3600
    stations = sorted({tr.id[:-1] for tr in stream})
    tasks = []
    for station1, station2 in itertools.combinations_with_replacement(
            stations, 2):
        if only_auto_correlation and station1 != station2:
            continue
        if station_combinations and not any(set(station_comb.split('-')) == (
                {station1.rsplit('.', 2)[0], station2.rsplit('.', 2)[0]}
                if '.' in (station_comb) else
                {station1.split('.')[1], station2.split('.')[1]})
                for station_comb in station_combinations):
            continue
        stream1 = Stream([tr for tr in stream if tr.id[:-1] == station1])
        stream2 = Stream([tr for tr in stream if tr.id[:-1] == station2])
        datetime1 = _midtime(stream1[0].stats)
        datetime2 = _midtime(stream2[0].stats)
        msg = 'Cannot get coordinates for channel %s datetime %s'
        try:
            c1 = inventory.get_coordinates(stream1[0].id, datetime=datetime1)
        except Exception as ex:
            raise RuntimeError(msg % (stream1[0].id, datetime1)) from ex
        try:
            c2 = inventory.get_coordinates(stream2[0].id, datetime=datetime2)
        except Exception as ex:
            raise RuntimeError(msg % (stream2[0].id, datetime2)) from ex
        args = (c1['latitude'], c1['longitude'],
                c2['latitude'], c2['longitude'])
        dist, azi, baz = gps2dist_azimuth(*args)
        if ('R' in components or 'T' in components) and station1 != station2:
            stream1 = stream1.copy()
            stream1b = stream1.copy().rotate('NE->RT', azi)
            stream1.extend(stream1b.select(component='R'))
            stream1.extend(stream1b.select(component='T'))
            stream2 = stream2.copy()
            stream2b = stream2.copy().rotate('NE->RT', azi)
            stream2.extend(stream2b.select(component='R'))
            stream2.extend(stream2b.select(component='T'))
        it_ = (itertools.product(stream1, stream2) if station1 != station2 else
               itertools.combinations_with_replacement(stream1, 2))
        for tr1, tr2 in it_:
            comps = tr1.stats.channel[-1] + tr2.stats.channel[-1]
            if component_combinations and (
                    comps not in component_combinations and
                    comps[::-1] not in component_combinations):
                continue
            tasks.append((tr1, tr2, dist, azi, baz))
    # start correlation
    do_work = partial(_slide_and_correlate_traces, day, next_day, length,
                      overlap, discard, max_lag, outkey,
                      demean_window)
    streams = start_parallel_jobs_inner_loop(tasks, do_work, njobs)
    xstream = Stream()
    xstream.traces = [tr for s_ in streams for tr in s_]
    if len(xstream) > 0:
        res = {}
        if keep_correlations:
            res['corr'] = xstream
        if stack:
            res['stack'] = yam.stack.stack(xstream, stack)
        return res



# Copyright 2017-2019 Tom Eulenfeld, MIT license

import unittest
from pkg_resources import load_entry_point
from contextlib import redirect_stderr, redirect_stdout
import glob
import io
import logging
import os
import shutil
import sys
import time
import tempfile
import tqdm

import matplotlib
matplotlib.use('Agg')


def _replace_in_file(fname_src, fname_dest, str_src, str_dest):
    with open(fname_src) as f:
        text = f.read()
    text = text.replace(str_src, str_dest)
    with open(fname_dest, 'w') as f:
        f.write(text)


class TestCase(unittest.TestCase):

    def setUp(self):
        args = sys.argv[1:]
        self.verbose = '-v' in args
        self.permanent_tempdir = '-p' in args
        self.less_data = '--full' not in args
        self.njobs = args[args.index('-n') + 1] if '-n'  in args else None
        if self.permanent_tempdir:
            tempdir = os.path.join(tempfile.gettempdir(), 'yam_test')
            if os.path.exists(tempdir) and '-d' in args:
                shutil.rmtree(tempdir)
            if not os.path.exists(tempdir):
                os.mkdir(tempdir)
        else:
            tempdir = tempfile.mkdtemp(prefix='yam_test_')
        self.cwd = os.getcwd()
        os.chdir(tempdir)
        self.tempdir = tempdir
        # for coverage put .coveragerc config file into tempdir
        # and append correct data_file parameter to config file
        covfn = os.path.join(self.cwd, '.coverage')
        if not os.path.exists('.coveragerc') and os.path.exists(covfn + 'rc'):
            _replace_in_file(covfn + 'rc', '.coveragerc', '[run]',
                             '[run]\ndata_file = ' + covfn)

        self.plotdir = os.path.join(self.tempdir, 'plots')
        self.script = load_entry_point('yam', 'console_scripts', 'yam')
        total = 74 - 2 * self.permanent_tempdir
        self.pbar = tqdm.tqdm(total=total, desc='CLI tests passed')

    def out(self, cmd, text=None):
        """Test if text is in output of command"""
        # for TRAVIS use maximal two cores
        if (self.njobs and '--njobs' not in cmd and
                cmd.split()[0] in ('correlate', 'stretch')):
            cmd = cmd + ' --njobs ' + self.njobs
        # disabling the logger is necessary, because the logging
        # configuration cannot be changed easily on subsequent calls
        # of yam in this test suite
        if self.verbose and cmd.split()[0] in ('correlate', 'stack', 'stretch'):
            if '-v' not in cmd:
                cmd = cmd + ' -vvv'
            logging.getLogger('yam').disabled = False
        elif self.verbose:
            logging.getLogger('yam').disabled = True
        if self.verbose:
            tqdm.tqdm.write('> yam ' + cmd)
        # catching all output, print only if tests are run with -v
        try:
            with io.StringIO() as f:
                with redirect_stdout(f), redirect_stderr(f):
                    try:
                        self.script(cmd.split())
                    except SystemExit:
                        pass
                output = f.getvalue()
            if self.verbose:
                tqdm.tqdm.write(output)
        finally:
            self.pbar.update(1)
        if text is not None:
            self.assertIn(text, output)
        return output

    def checkplot(self, bname):
        """Test if plot file exists"""
        fname = os.path.join(self.plotdir, bname)
        self.assertTrue(os.path.exists(fname), msg='%s missing' % fname)

    def test_cli(self):
        # create tutorial
        self.out('-h', 'print information about')
        self.out('--version', '.')
        self.out('bla', 'invalid choice')
        if not self.permanent_tempdir:
            self.out('info', "No such file or directory: 'conf.json'")
        self.out('create', '')
        if not self.permanent_tempdir:
            self.out('info', 'Not found')
        cmd = 'create --tutorial' + ' --less-data' * self.less_data
        self.out(cmd)
        self.out(cmd)
        self.out('scan', 'obspy-scan')

        # check basics
        _replace_in_file('conf.json', 'conf2.json', '"io"', '"io",')
        self.out('-c conf2.json info', 'parsing the conf')
        self.out('info', '3 stations')
        self.out('info stations', 'CX.PATCX..BHZ')
        self.out('info data', 'example_data/CX.PATCX')
        self.out('print stations', 'CX.PATCX..BHZ')
        pr = self.out('print data CX.PB06..BHZ 2010-02-05', '2010-02-05')
        self.out('print data CX.PATCX..BHZ 2010-036', '2010-02-05')
        self.out('print prepdata CX.PATCX..BHZ 2010-02-05 1', 'samples')
        self.out('print data', 'seedid')
        self.out('print prepdata CX.PATCX..BHZ 2010-02-05', 'corrid')
        # check remove_response on one day
        _replace_in_file(
            'conf.json', 'conf2.json', '{"clip_factor": 2},',
            '{"clip_factor":2, "clip_set_zero":true},"remove_response":true,')
        self.out('-c conf2.json print prepdata CX.PATCX..BHZ 2010-02-05 1a')

        # comment until ObsPy issue 2097 is fixed
#        try:
#            self.out('plot stations')  # takes long
#        except ImportError:
#            pass
#        else:
#            self.checkplot('stations.png')
        self.out('plot data CX.PATCX..BHZ 2010-02-05')
        self.checkplot('data_CX.PATCX..BHZ_2010-02-05.png')
        self.out('plot prepdata CX.PATCX..BHZ 2010-02-05 1')
        self.checkplot('prepdata_CX.PATCX..BHZ_2010-02-05_c1.png')

        # check data_plugin
        data_plugin_file = """
from obspy import read
EXPR = ("example_data/{network}.{station}.{location}.{channel}__"
        "{t.year}{t.month:02d}{t.day:02d}*.mseed")
# def get_data(starttime, endtime, network, station, location, channel):
def get_data(starttime, endtime, **smeta):
     fname = EXPR.format(t=starttime, **smeta)
     stream = read(fname, 'MSEED')
     # load also previous file, because startime in day file is after 0:00
     fname = EXPR.format(t=starttime-5, **smeta)
     stream += read(fname, 'MSEED')
     return stream
        """
        with open('data.py', 'w') as f:
            f.write(data_plugin_file)
        _replace_in_file('conf.json', 'conf2.json', '"data_plugin": null',
                         '"data_plugin": "data : get_data"')
        self.out('-c conf2.json info', 'data : get_data')
        self.out('-c conf2.json info data', 'data : get_data')
        pr2 = self.out('-c conf2.json print data CX.PB06..BHZ 2010-02-05')
        self.assertEqual(pr, pr2)

        # check correlation
        t1 = time.time()
        self.out('correlate 1')  # takes long
        t2 = time.time()
        self.out('correlate 1 -v')
        t3 = time.time()
        if not self.permanent_tempdir:
            self.assertLess(t3 - t2, 0.5 * (t2 - t1))
        self.out('correlate auto')  # takes long
        self.out('correlate 1a')  # takes long
        self.out('info', 'c1_s1d: 7 combs')
        self.out('info', 'c1a_s1d: 3 combs')
        self.out('info', 'cauto: 2 combs')
        cauto_info = self.out('info cauto', 'CX.PATCX/.BHZ-.BHZ/2010-02-05')
        self.out('info c1_s1d/CX.PB06-CX.PB06/.BHZ-.BHZ', 'CX.PB06-CX.PB06')
        self.out('print cauto', '1201 samples')
        expected = '%d Trace' % (2 if self.less_data else 11)
        self.out('print c1_s1d/CX.PB06-CX.PB06/.BHZ-.BHZ', expected)
        # check if correlation without parallel processing gives the same
        # result keys
        with self.assertWarnsRegex(UserWarning, 'only top level keys'):
            self.out('remove cauto/CX.PATCX-CX.PATCX')
        self.out('remove cauto')
        self.out('correlate auto --parallel-inner-loop')  # takes long
        self.maxDiff = None
        cauto_info_seq = self.out('info cauto')
        self.assertEqual(cauto_info, cauto_info_seq)

        # check plots of correlation
        po = ('--plot-options {"xlim":[0,10],"figsize":[10,10],'
                              '"ylim":[null,"2010-02-05"]}')
        self.out('plot c1_s1d --plottype vs_dist')
        self.checkplot('corr_vs_dist_c1_s1d_ZZ.png')
        self.out('plot c1_s1d/CX.PATCX-CX.PB06 --plottype wiggle')
        self.out('plot cauto --plottype wiggle %s' % po)
        bname = 'corr_vs_time_wiggle_c1_s1d_CX.PATCX-CX.PB06_'
        self.checkplot(bname + '.BHZ-.BHZ.png')
        self.checkplot(bname + '.BHN-.BHZ.png')
        bname = 'corr_vs_time_wiggle_cauto_CX.PATCX-CX.PATCX_'
        self.checkplot(bname + '.BHZ-.BHZ.png')
        self.checkplot(bname + '.BHN-.BHZ.png')
        self.out('plot c1_s1d/CX.PATCX-CX.PB06')
        self.out('plot cauto %s' % po)
        bname = 'corr_vs_time_c1_s1d_CX.PATCX-CX.PB06_'
        self.checkplot(bname + '.BHZ-.BHZ.png')
        self.checkplot(bname + '.BHN-.BHZ.png')
        bname = 'corr_vs_time_cauto_CX.PATCX-CX.PATCX_'
        self.checkplot(bname + '.BHZ-.BHZ.png')
        self.checkplot(bname + '.BHN-.BHZ.png')

        # check stacking
        self.out('stack c1_s1d 2d')
        self.out('info',  'c1_s1d_s2d: 7 combs')
        self.out('stack c1_s1d 2dm1d')
        self.out('info',  'c1_s1d_s2dm1d: 7 combs')
        self.out('stack c1_s1d 1')
        self.out('stack cauto 2')  # takes long
        self.out('info',  'cauto_s2')
        self.out('remove c1_s1d_s2dm1d c1_s1d_s1')

        # check stretching
        po = ('--plot-options {"show_line":true,'
                              '"xlim":[null,"2010-02-05"]}')
        self.out('stretch c1_s1d/CX.PATCX-CX.PATCX 1')
        self.out('stretch c1_s1d/CX.PATCX-CX.PATCX 1b')
        self.out("stack c1_s1d None")
        self.out('stretch c1_s1d 1 --reftrid c1_s1d_s')
        self.out('stretch cauto/CX.PATCX-CX.PATCX/.BHZ-.BHZ 2')
        self.out('stretch cauto 2')
        self.out('remove cauto_t2/CX.PATCX-CX.PATCX/.BHZ-.BHZ')
        self.out('stretch cauto 2 --njobs 1')
        self.out('remove cauto_t2/CX.PATCX-CX.PATCX/.BHZ-.BHZ')
        _replace_in_file('conf.json', 'conf2.json', '"sides": "right"',
                '"time_period": [null, "2010-02-05"], "max_lag": 40, '
                '"sides": "right"')
        self.out('-c conf2.json stretch cauto/CX.PATCX-CX.PATCX/.BHZ-.BHZ 2')
        self.out('-c conf2.json stretch cauto/CX.PATCX-CX.PATCX/.BHZ-.BHZ 2b')
        self.out('plot c1_s1d_t1/CX.PATCX-CX.PB06 %s' % po)
        self.out('plot c1_s1d_t1b/CX.PATCX-CX.PB06 %s' % po)
        self.out('plot cauto_t2 --plottype wiggle', 'not supported')
        self.out('plot cauto_t2')
        self.out('plot cauto_t2b')
        globexpr = os.path.join(self.plotdir, 'sim_mat_cauto_t2*.png')
        self.assertEqual(len(glob.glob(globexpr)), 3)
        globexpr = os.path.join(self.plotdir, 'sim_mat_c1_s1d_t1*.png')
        if not self.permanent_tempdir:
            self.assertEqual(len(glob.glob(globexpr)), 3)
        self.out('plot c1_s1d_t1')
        num_plots = 5 if self.less_data else 7
        self.assertEqual(len(glob.glob(globexpr)), num_plots)
        num_combs = 5 if self.less_data else 7
        self.out('info', 'c1_s1d_t1: %d combs' % num_combs)
        self.out('info cauto_t2', 'CX.PATCX-CX.PATCX/.BHZ-.BHZ')
        info_t = self.out('print cauto_t2', 'num_stretch')
        self.assertGreater(len(info_t.splitlines()), 100)  # lots of lines

        # check export
        fname = 'dayplot.mseed'
        self.out('export data %s CX.PB06..BHZ 2010-02-05' % fname)
        self.assertTrue(os.path.exists(fname), msg='%s missing' % fname)
        fname = 'dayplot.mseed'
        self.out('export data %s CX.PB06..BHZ 2010-02-05' % fname)
        self.assertTrue(os.path.exists(fname), msg='%s missing' % fname)
        fname = 'some_auto_corrs.h5'
        self.out('export cauto/CX.PATCX-CX.PATCX %s --format H5' % fname)
        self.assertTrue(os.path.exists(fname), msg='%s missing' % fname)

        # check load (IPython mocked)
        sys.modules['IPython'] = unittest.mock.MagicMock()
        self.out('load cauto', 'Good Bye')
        self.out('load c1_s1d', 'Good Bye')
        self.out('load c1_s1d_t1/CX.PATCX-CX.PB01', 'Good Bye')

    def tearDown(self):
        os.chdir(self.cwd)
        if not self.permanent_tempdir:
            shutil.rmtree(self.tempdir)


def suite():
    return unittest.makeSuite(TestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# Copyright 2017-2018 Tom Eulenfeld, GPLv3
import os.path
import re

from setuptools import find_packages, setup


def find_version(*paths):
    fname = os.path.join(os.path.dirname(__file__), *paths)
    with open(fname) as fp:
        code = fp.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", code, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


VERSION = find_version('yam', '__init__.py')
DESCRIPTION = (
    'Yet another monitoring tool using correlations of '
    'ambient noise (seismology)')
LONG_DESCRIPTION = (
    'Please look at the project site for tutorials and information.')

ENTRY_POINTS = {
    'console_scripts': ['yam-runtests = yam.tests:run',
                        'yam = yam.main:run_cmdline']}

REQUIRES = ['h5py', 'matplotlib', 'numpy', 'obspy>=1.1', 'obspyh5>=0.3',
            'scipy>=0.18', 'setuptools', 'tqdm']

CLASSIFIERS = [
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering :: Physics'
    ]


setup(name='yam',
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      url='https://github.com/trichter/yam',
      author='Tom Eulenfeld',
      author_email='tom.eulenfeld@gmail.de',
      license='MIT',
      packages=find_packages(),
      package_dir={'yam': 'yam'},
      install_requires=REQUIRES,
      entry_points=ENTRY_POINTS,
      include_package_data=True,
      zip_safe=False,
      classifiers=CLASSIFIERS
      )



# Copyright 2017-2019 Tom Eulenfeld, MIT license
"""Reading and writing correlations and stretching results"""

import h5py

import obspyh5
import yam
import os.path
import obspy

from yam.util import _analyze_key, _get_fname


INDEX = ('{key}/{network1}.{station1}-{network2}.{station2}/'
         '{location1}.{channel1}-{location2}.{channel2}/'
         '{starttime.datetime:%Y-%m-%dT%H:%M}')
INDEX_STRETCH = ('{key}/{network1}.{station1}-{network2}.{station2}/'
                 '{location1}.{channel1}-{location2}.{channel2}')
obspyh5.set_index(INDEX)


def write_dict(dict_, fname, mode='a', libver='latest', dtype='float16'):
    """
    Write similarity matrix into HDF5 file

    :param dict_: Dictionary with stretching results
        (output from `~yam.stretch.stretch()`)
    :param fname: file name
    :param mode: file mode (default ``'a'`` -- write into file)
    :param libver: use latest version of HDF5 file format
    :param dtype: type for similarity matrix
    """
    with h5py.File(fname, mode=mode, libver=libver) as f:
        f.attrs['file_format_stretch'] = 'yam'
        f.attrs['version_stretch'] = yam.__version__
        if 'index_stretch' not in f.attrs:
            index = f.attrs['index_stretch'] = INDEX_STRETCH
        else:
            index = f.attrs['index_stretch']
        attrs = dict_['attrs']
        index = index.format(**attrs)
        group = f.require_group(index)
        for key, val in attrs.items():
            if 'time' in key:
                val = str(val)
            group.attrs[key] = val
        group.create_dataset('sim_mat', data=dict_['sim_mat'], dtype=dtype)
        for key, val in dict_.items():
            if key not in ('attrs', 'sim_mat'):
                group.create_dataset(key, data=val)


def _get_existent(fname, root, level):
    """
    Return existing keys at level in HDF5 file
    """
    if not os.path.exists(fname):
        return []
    done = []

    def visit(group, level):
        if level == 0:
            done.append(group.name)
            return
        elif level == 1:
            done.extend([group.name + '/' + subg for subg in group])
            return
        for n in group:
            visit(group[n], level - 1)
    with h5py.File(fname, 'r') as f:
        try:
            g = f[root]
        except KeyError:
            return []
        if root == '/':
            level_reached = 0
        else:
            level_reached = g.name.count('/')
        visit(g, level - level_reached)
    return sorted(done)


def _read_dict(group, readonly=None):
    """Read a single stretching dictionary from group"""
    res = {'attrs': {}}
    for key, val in group.attrs.items():
        res['attrs'][key] = val
        if key in ('starttime', 'endtime'):
            res['attrs'][key] = obspy.UTCDateTime(val)
    for key, val in group.items():
        if key != 'attrs' and (readonly is None or key in readonly):
            res[key] = val[:]
#    if 'times' in res:
#        res['times'] = [UTC(t) for t in res['times']]
    res['group'] = group.name
    return res


def _iter_dicts(fname, groupname='/', level=3, readonly=None):
    """Iterator yielding stretching dictionaries"""
    tasks = _get_existent(fname, groupname, level)
    with h5py.File(fname, 'r') as f:
        for task in tasks:
            yield task, _read_dict(f[task], readonly=readonly)


def read_dicts(fname, groupname='/', level=3, readonly=None):
    """
    Read dictionaries with stretching results

    :param fname: file name
    :param groupname: specify group to read
    :param level: level in index where the data was written, defaults to 3 and
        should not be changed
    :return: list of dictionaries with stretching results
    """
    return [obj[1] for obj in _iter_dicts(fname, groupname, level, readonly)]


def _iter_streams(fname, groupname='/', level=3):
    """Iterator yielding correlation streams"""
    tasks = _get_existent(fname, groupname, level)
    for task in tasks:
        stream = obspy.read(fname, 'H5', group=task)
        yield task, stream


def _iter_h5(io, key, level=3):
    """Iterator yielding streams or stretching results, depending on key"""
    is_stretch = 't' in _analyze_key(key)
    fname = _get_fname(io, key)
    iter_ = _iter_dicts if is_stretch else _iter_streams
    for obj in iter_(fname, key, level=level):
        yield obj


def _write_corr(result, io, dtype=None):
    """Write result from yam.correlate.correlate"""
    if result is not None:
        for key in result:
            result[key].write(io[key], 'H5', dtype=dtype, mode='a')



# Copyright 2017-2018 Tom Eulenfeld, GPLv3
"""
yam Documentation
=================


Motivation
----------

Why another monitoring tool for seismic velocities using ambient noise cross-correlations?

There are several alternatives around, namely `MSNoise <https://github.com/ROBelgium/MSNoise>`_
and MICC `MIIC <https://github.com/miic-sw/miic>`_.
*MSNoise* is especially useful for large datasets and continuous monitoring. Configuration and the state of a project
is managed by sqlite or mysql database. A project can be configured via web interface,
commands are issued via command line interface. Velocity variations are determined with the
Moving Window Cross Spectral technique (MWCS).
*MIIC* is another monitoring library using the time-domain stretching technique.

*Yam*, contrary to MSNoise, is designed for off-line usage, but also includes capabilities to reprocess continuously
growing data. Yam does not rely onto a database, but rather checks on the fly which results already exist and which
results have still to be calculated.
Cross-correlations are written to HDF5 files via the ObsPy plugin obspyh5. Thus, correlation data can be easily
accessed with ObsPy's |read| function after the calculation. It follows a similar processing flow as MSNoise,
but it uses the stretching library from MIIC. (It is of course feasible to implement MWCS.)
One of its strong points is the configuration declared in a simple, but heavily commented JSON file.
It is possible to declare similar configurations.
A possible use case is the reprocessing of the whole dataset in a different frequency band.
Some code was reused from previous project `sito <https://github.com/trichter/sito>`_.


Installation
------------

Dependencies of yam are ``obspy>=1.1 obspyh5>=0.3 h5py``.
Optional dependencies are ``IPython`` and ``cartopy``.
The recommended way to install yam is via `anaconda <https://docs.anaconda.com/anaconda/install.html>`_ and pip::

    conda --add channels conda-forge
    conda create -n yam cartopy h5py IPython matplotlib numpy obspy scipy tqdm
    source activate yam
    pip install yam

After that, you can run the tests with ``yam-runtests`` and check if everything is installed properly.


How to use yam
--------------

The scripts are started with the command line program ``yam``.
``yam -h`` gives an overview over available commands and options. Each command has its own help,
e.g. ``yam correlate -h`` will print help for the ``correlate`` command.

``create`` will create an example configuration file in JSON format.
The processing commands are ``correlate``, ``stack`` and ``stretch``.

``info``, ``print``, ``load`` and ``plot`` commands allow to inspect correlations,
stacks and stretching results as well as preprocessed data and other aspects.
``remove`` removes correlations or stretching results (necessary if configuration changed).

Correlations, corresponding stacks and stretching results are saved in HDF5 files.
The indices inside the HDF5 files are the following (first for correlations, second for stretching results)::

    '{key}/{network1}.{station1}-{network2}.{station2}/{location1}.{channel1}-{location2}.{channel2}/{starttime.datetime:%Y-%m-%dT%H:%M}'
    '{key}/{network1}.{station1}-{network2}.{station2}/{location1}.{channel1}-{location2}.{channel2}'

The strings are expanded with the corresponding metadata.
Several tools are available for analysing the contents of the HDF5 files, e.g. h5ls or hdfview.


About keys and different configurations
***************************************

``key`` in the above indices and as a parameter in the command line interface
is a special parameter which describes the processing chain.
It is best explained with an example: A key could be ``c1_s2d_twow``.
This means data was correlated (``c``) with configuration ``1``, each two days ``2d`` are stacked (``s``) and
finally data was stretched (``t``) using the stretching configuration ``wow``.
The configuration of the keys are defined in the configuration file.
(These ids may not contain ``_``, because ``_`` is used to separate the different processing steps.)
The ``s`` key is special, because it can describe the stacking procedure directly:
For example, ``s5d`` stacks correlations of 5 days,
``s2h`` of 2 hours, ``s5dm2.5d`` is a 2.5 day moving (``m``) stack over 5 days,
with ``d`` corresponding to days and ``h`` corresponding to hours.
But ``s`` can also precede a key which is described in the configuration file.

Valid processing chains could be represented by ``c2`` (data is only correlated),
``c2_t2`` (and directly stretched afterwards),
``c1_s10dm5d_t1`` (correlation, moving stack, stretch),
``c1_s1d_s5dm2d`` (correlation, stack, moving stack) or similar.


Tutorial
********

A small tutorial with an example dataset is included.
It can be loaded into an empty directory with ``yam create --tutorial``.
Now you can try out some of the following commands:

.. code:: bash

    yam info               # plot information about project
    yam info stations      # print inventory info
    yam info data          # plot info about data files
    yam plot stations      # plot station map
    yam print data CX.PATCX..BHZ 2010-02-03       # load data for a specific station and day and print information
    yam load data CX.PATCX..BHZ 2010-02-03        # load data for a specific station and day and start an IPython session
    yam plot data CX.PATCX..BHZ 2010-02-03        # plot a day file
    yam plot prepdata CX.PATCX..BHZ 2010-02-03 1  # plot the preprocessed data of the same day
                                                  # (preprocessing defined in corr config 1)
    yam correlate 1        # correlates data with corr configuration 1
    yam correlate 1        # should finish fast, because everything is already calculated
    yam correlate auto     # correlate data with another configuration suitable for auto-correlations
    yam plot c1_s1d --plottype vs_dist  # plot correlation versus distance
    yam plot cauto --plot-options '{"xlim": [0, 10]}'  # plot auto-correlations versus time and change some options
                                                       # ("wiggle" plot also possible)
    yam stack c1_s1d 3dm1d       # stack 1 day correlations with a moving stack of 3 days
    yam stack cauto 2            # stack auto-correlations with stack configuration 2

    yam stretch c1_s1d_s3dm1d 1  # stretch the stacked data with stretch configuration 1
    yam stretch cauto_s2 2       # stretch the stacked auto-correlations with another stretch configuration
    yam info                     # find out about the keys which are already in use
    yam plot cauto_t2            # plot similarity matrices for the given processing chain
    yam plot cauto_s2_t2 --plot-options '{"show_line": true}' --show  # plot all similarity matrices for this processing
                                                                      # chain and display them on screen (zoom etc.)
    yam plot c1_s1d_s3dm1d_t1/CX.PATCX-CX.PB01  # plot similarity matrices, but only for one station combination
                                                # (restricting the group is also possible for stacking and stretching)

Of course, the plots do not look overwhelmingly for such a small dataset.

The readme in the Github repository links to a further advanced tutorial.


Use your own data
*****************

Create the example configuration with ``yam create`` and adapt it to your needs.
A good start is to change the ``inventory`` and ``data`` parameters.


Read correlation and results of stretching procedure in Python for further processing
*************************************************************************************

Use ObsPy's |read| to read correlations and stacks and `~yam.commands.read_dicts()` to read stretching results.

::

    from obspy import read
    from yam import read_dicts

    # read a whole file of correlations
    stream = read('corr.h5', 'H5')
    # to read only part of a file
    stream = read('stack.h5', 'H5', include_only=dict(key='c1_s1d', network1='CX', station1='PATCX',
                                                      network2='CX', station2='PB01'))
    # or specify the group explicitly
    stream = read('stack.h5', 'H5', group='c1_s1d')
    # read the stretching results into a dictionary
    stretch_result = read_dicts('stretch.h5', 'c1_s1d_t1')


Configuration options
*********************

Please see the :ref:`example configuration file <config_label>` configuration file
for an explanation of configuration options.
It follows a table with links to functions which consume the options.
All config options should be documented inside these functions.

========================  ========================================
configuration dictionary      functions consuming the options
========================  ========================================
    io                    Configuration for input and output (needed by most functions in `~yam.commands` module)
    correlate             `~yam.commands.start_correlate` -> `~yam.correlate.correlate`  ->  `~yam.correlate.preprocess`  -> `~yam.correlate.time_norm`, `~yam.correlate.spectral_whitening`
    stack                 `~yam.commands.start_stack` -> `~yam.stack.stack`
    stretch               `~yam.commands.start_stretch` -> `~yam.commands.stretch_wrapper` -> `~yam.stretch.stretch`
    plot_*_options        See corresponding functions in `~yam.imaging` module
========================  ========================================

More information about the different subcommands of yam can be found in the corresponding functions in
`~yam.commands` module.

.. shortcuts e.g. for references to Obspy
.. |read| replace:: :func:`~obspy.core.stream.read`
.. |UTC| replace:: `~obspy.core.utcdatetime.UTCDateTime`
.. |Trace| replace:: `~obspy.core.trace.Trace`
.. |Stream| replace:: `~obspy.core.stream.Stream`
.. |Stream.plot| replace:: `Stream.plot() <obspy.core.stream.Stream.plot>`
.. |Inventory| replace:: `~obspy.core.inventory.inventory.Inventory`
.. |Inventory.plot| replace:: `Inventory.plot() <obspy.core.inventory.inventory.Inventory.plot>`
.. |Inventory.select| replace:: `Inventory.filter() <obspy.core.inventory.inventory.Inventory.select>`
.. |Axes.plot| replace:: `Axes.plot() <matplotlib.axes.Axes.plot>`

.. |io| replace:: io configuration dictionary
"""

__version__ = '0.5.0'

from yam.main import run
from yam.commands import read_dicts



"""
Tests for the rf package.

yam-runtests [-h] [-v] [-p] [-d] [-f] [-n num]

-h    short help
-v    be verbose
-p    use permanent tempdir
-d    empty permanent tempdir at start
--full   use full tutorial dataset (tests take longer)
-n num   maximal number of cores to use (default: all)
"""

from pkg_resources import resource_filename
import sys
import unittest


def run():
    if '-h' in sys.argv[1:]:
        print(__doc__)
        return
    loader = unittest.TestLoader()
    test_dir = resource_filename('yam', 'tests')
    suite = loader.discover(test_dir)
    runner = unittest.runner.TextTestRunner()
    ret = not runner.run(suite).wasSuccessful()
    sys.exit(ret)
